class BaseError(Exception):
    def __init__(self, token, message):
        self.token   = token
        self.message = message


class Return(Exception):
    def __init__(self, value):
        self.value = value


from hyde.expressions import Visitor as ExpressionVisitor
from hyde.statements import Visitor as StatementVisitor


class Visitor(ExpressionVisitor, StatementVisitor):
    def visit(self, other):
        # This should never happen, but prevent Python errors from showing if it does
        if other is None:
            return

        visit_fn_name = f'visit_{type(other).__name__.lower()}'
        visit_fn = getattr(self, visit_fn_name)

        return visit_fn(other)


class Statement:
    pass


class Block(Statement):
    def __init__(self, statements):
        self.statements = statements


class ClassDef(Statement):
    def __init__(self, name, superclass, methods):
        self.name = name
        self.superclass = superclass
        self.methods = methods


class Expression(Statement):
    def __init__(self, expression):
        self.expression = expression


class Function(Statement):
    def __init__(self, name, params, body):
        self.name = name
        self.params = params
        self.body = body


class IfStmt(Statement):
    def __init__(self, condition, then_branch, else_branch):
        self.condition = condition
        self.then_branch = then_branch
        self.else_branch = else_branch


class Load(Statement):
    def __init__(self, path):
        self.path = path


class Print(Statement):
    def __init__(self, expression):
        self.expression = expression


class ReturnStmt(Statement):
    def __init__(self, keyword, value):
        self.keyword = keyword
        self.value = value


class Var(Statement):
    def __init__(self, name, initializer):
        self.name = name
        self.initializer = initializer


class WhileStmt(Statement):
    def __init__(self, condition, body):
        self.condition = condition
        self.body = body


class Visitor:
    # Base methods to be overridden in child classes
    def visit_block(self, block):
        raise NotImplementedError('visit_block')

    def visit_classdef(self, classdef):
        raise NotImplementedError('visit_classdef')

    def visit_expression(self, expression):
        raise NotImplementedError('visit_expression')

    def visit_function(self, function):
        raise NotImplementedError('visit_function')

    def visit_ifstmt(self, ifstmt):
        raise NotImplementedError('visit_ifstmt')

    def visit_load(self, load):
        raise NotImplementedError('visit_load')

    def visit_print(self, print):
        raise NotImplementedError('visit_print')

    def visit_returnstmt(self, returnstmt):
        raise NotImplementedError('visit_returnstmt')

    def visit_var(self, var):
        raise NotImplementedError('visit_var')

    def visit_whilestmt(self, whilestmt):
        raise NotImplementedError('visit_whilestmt')




from hyde.environment import Environment
from hyde.errors import BaseError


class HydeInstanceError(BaseError):
    pass


class HydeInstance:
    def __init__(self, klass):
        self.klass  = klass
        self.fields = {}
    
    def get(self, name):
        if name.lexeme in self.fields:
            return self.fields[name.lexeme]

        method = self.klass.find_method(name.lexeme)
        if method is not None:
            return method.bind(self)

        raise HydeInstanceError(name, f'Undefined property {name.lexeme} for {self.klass} instance.')

    def set(self, name, value):
        self.fields[name.lexeme] = value

    def __str__(self):
        return f'{self.klass.name} instance'


from hyde.errors import BaseError


class RuntimeError(BaseError):
    pass


class Environment:
    def __init__(self, enclosing = None):
        self.values    = {}
        self.enclosing = enclosing

    def assign(self, name, value):
        if name.lexeme in self.values:
            self.values[name.lexeme] = value
            return

        if self.enclosing is not None:
            return self.enclosing.assign(name, value)

        raise RuntimeError(name, f'Undefined variable {name.lexeme}.')

    def assign_at(self, distance, name, value):
        self.ancestor(distance).values[name.lexeme] = value

    def define(self, name, value):
        self.values[name] = value

    def get(self, name):
        if name.lexeme in self.values:
            return self.values[name.lexeme]

        if self.enclosing is not None:
            return self.enclosing.get(name)

        raise RuntimeError(name, f'Undefined variable {name.lexeme}.')
    
    def get_at(self, distance, name):
        return self.ancestor(distance).values.get(name)

    def ancestor(self, distance):
        environment = self

        while distance > 0:
            distance -= 1
            environment = environment.enclosing

        return environment

    def merge(self, other):
        target = self

        while other:
            target.values.update(other.values)

            target = self.enclosing
            other  = other.enclosing



from enum import Enum, auto, unique
from hyde.errors import BaseError
from hyde.visitor import Visitor


@unique
class FunctionType(Enum):
    NONE        = auto()
    FUNCTION    = auto()
    INITIALIZER = auto()
    METHOD      = auto()


@unique
class ClassType(Enum):
    NONE     = auto()
    CLASS    = auto()
    SUBCLASS = auto()


class ResolutionError(BaseError):
    pass


class Resolver(Visitor):
    def __init__(self, interpreter):
        self.interpreter   = interpreter
        self.scopes        = [{}]
        self.current_fn    = FunctionType.NONE
        self.current_class = ClassType.NONE

    # Visit Expressions
    def visit_assign(self, expr):
        self.resolve_single(expr.value)
        self.resolve_local(expr, expr.name)

    def visit_binary(self, expr):
        self.resolve_single(expr.left)
        self.resolve_single(expr.right)

    def visit_call(self, expr):
        self.resolve_single(expr.callee)

        for argument in expr.arguments:
            self.resolve_single(argument)

    def visit_get(self, expr):
        self.resolve_single(expr.object)

    def visit_grouping(self, expr):
        self.resolve_single(expr.expression)

    def visit_literal(self, expr):
        pass

    def visit_logical(self, expr):
        self.resolve_single(expr.left)
        self.resolve_single(expr.right)

    def visit_set(self, expr):
        self.resolve_single(expr.value)
        self.resolve_single(expr.object)

    def visit_super(self, expr):
        if self.current_class == ClassType.NONE:
            raise ResolutionError(expr.keyword, "Can't use 'super' outside of a class.")
        elif self.current_class != ClassType.SUBCLASS:
            raise ResolutionError(expr.keyword, "Can't use 'super' in a class with no superclass.")
        self.resolve_local(expr, expr.keyword)

    def visit_this(self, expr):
        if self.current_class == ClassType.NONE:
            raise ResolutionError(expr.keyword, "Can't use 'this' outside of a class.")

        self.resolve_local(expr, expr.keyword)

    def visit_unary(self, expr):
        self.resolve_single(expr.right)

    def visit_variable(self, expr):
        if len(self.scopes) != 0 and self.scopes[-1].get(expr.name.lexeme) == False:
            raise ResolutionError(expr.name, "Can't read local variable in its own initializer.")
    
        self.resolve_local(expr, expr.name)

    # Visit Statements
    def visit_block(self, stmt):
        self.begin_scope()
        self.resolve(stmt.statements)
        self.end_scope()

    def visit_classdef(self, stmt):
        enclosing_class = self.current_class
        self.current_class = ClassType.CLASS

        self.declare(stmt.name)
        self.define(stmt.name)

        if stmt.superclass is not None and stmt.name.lexeme == stmt.superclass.name.lexeme:
            raise ResolutionError(stmt.superclass.name, "A class can't inherit from itself.")

        if stmt.superclass is not None:
            self.current_class = ClassType.SUBCLASS
            self.resolve_single(stmt.superclass)

        if stmt.superclass is not None:
            self.begin_scope()
            self.scopes[-1]['super'] = True

        self.begin_scope()
        self.scopes[-1]['this'] = True

        for method in stmt.methods:
            declaration = FunctionType.METHOD

            if method.name.lexeme == 'init':
                declaration = FunctionType.INITIALIZER

            self.resolve_function(method, declaration)
        
        self.end_scope()

        if stmt.superclass is not None:
            self.end_scope()

        self.current_class = enclosing_class

    def visit_expression(self, stmt):
        self.resolve_single(stmt.expression)

    def visit_function(self, stmt):
        self.declare(stmt.name)
        self.define(stmt.name)

        self.resolve_function(stmt, FunctionType.FUNCTION)

    def visit_ifstmt(self, stmt):
        self.resolve_single(stmt.condition)
        self.resolve_single(stmt.then_branch)

        if stmt.else_branch is not None:
            self.resolve_single(stmt.else_branch)

    def visit_load(self, stmt):
        pass

    def visit_print(self, stmt):
        self.resolve_single(stmt.expression)

    def visit_returnstmt(self, stmt):
        if (self.current_fn == FunctionType.NONE):
            raise ResolutionError(stmt.keyword, "Can't return from top-level code.")

        if stmt.value is not None:
            if self.current_fn == FunctionType.INITIALIZER:
                raise ResolutionError(stmt.keyword, "Can't return a value from an initializer.")
            self.resolve_single(stmt.value)

    def visit_var(self, stmt):
        self.declare(stmt.name)
        if stmt.initializer is not None:
            self.resolve_single(stmt.initializer)

        self.define(stmt.name)

    def visit_whilestmt(self, stmt):
        self.resolve_single(stmt.condition)
        self.resolve_single(stmt.body)

    # Helpers
    def begin_scope(self):
        self.scopes.append({})

    def end_scope(self):
        self.scopes.pop()

    def resolve(self, statements):
        for statement in statements:
            self.resolve_single(statement)

    def resolve_single(self, other):
        self.visit(other)

    def resolve_local(self, expr, name):
        distance = None

        for i in range(len(self.scopes)):
            if name.lexeme in self.scopes[i]:
                distance = len(self.scopes) - 1 - i
        
        if distance is not None:
            self.interpreter.resolve(expr, distance)
        
    def resolve_function(self, function, fn_type):
        enclosing_fn    = self.current_fn
        self.current_fn = fn_type

        self.begin_scope()

        for param in function.params:
            self.declare(param)
            self.define(param)

        self.resolve(function.body)

        self.end_scope()

        self.current_fn = enclosing_fn

    def declare(self, name):
        if len(self.scopes) == 0:
            return

        scope = self.scopes[-1]

        if name.lexeme in scope:
            raise ResolutionError(name, 'Already a variable with this name in this scope.')

        scope[name.lexeme] = False

    def define(self, name):
        if len(self.scopes) == 0:
            return

        self.scopes[-1][name.lexeme] = True


from hyde.primitives.array import Array as HydeArray
from hyde.primitives.http import BasicHttpRequestHandler as HydeBasicHttpRequestHandler
from hyde.primitives.map import Map as HydeMap
from hyde.hyde_callable import HydeCallable


class Globals:
    class GlobalHydeCallable(HydeCallable):
        capitalized = False

        def __str__(self):
            return f'<fn {self.__name__}>'


    class Array(GlobalHydeCallable):
        arity = 1
        capitalized = True

        def call(self, _interpreter, arguments):
            size = int(arguments[0])

            return HydeArray(size)


    class BasicHttpRequestHandler(GlobalHydeCallable):
        arity = 1
        capitalized = True

        def call(self, _interpreter, arguments):
            request_attrs = arguments[0]

            return HydeBasicHttpRequestHandler(request_attrs)

    class Clock(GlobalHydeCallable):
        from time import time

        arity = 0

        def call(self, _interpreter, _arguments):
            return round(self.time() * 1000)

    
    class Map(GlobalHydeCallable):
        arity = 0
        capitalized = True

        def call(self, _interpreter, _arguments):
            return HydeMap()


    GLOBAL_FUNCTIONS = [
        Array,
        BasicHttpRequestHandler,
        Clock,
        Map
    ]

    @classmethod
    def define(cls, environment):
        for fn in cls.GLOBAL_FUNCTIONS:
            name = fn.__name__.lower() if not fn.capitalized else fn.__name__
            environment.define(name, fn())


from hyde.hyde_callable import HydeCallable
from hyde.hyde_instance import HydeInstance, HydeInstanceError
from hyde.primitives.array import Array


class Map(HydeInstance):
    name = "Map"

    class MapFunction(HydeCallable):
        def __init__(self, map, token):
            self.map   = map
            self.token = token


    class Get(MapFunction):
        arity = 1

        def call(self, _interpreter, arguments):
            key = arguments[0]

            return self.map.value_map.get(key)

        
    class Set(MapFunction):
        arity = 2

        def call(self, _interpreter, arguments):
            key = arguments[0]
            value = arguments[1]
            self.map.value_map[key] = value

            return value

        
    class Keys(MapFunction):
        arity = 0
        
        def call(self, _interpreter, _arguments):
            array = Array(len(self.map.value_map))
            array.elements = self.map.value_map.keys()

            return array


    def __init__(self):
        self.value_map = {}

    def get(self, name):
        if name.lexeme == 'get':
            return Map.Get(self, name)
        elif name.lexeme == 'set':
            return Map.Set(self, name)
        elif name.lexeme == 'keys':
            return Map.Keys(self, name)

        raise HydeInstanceError(name, f'Undefined property {name.lexeme} for {self.name} instance.')

    def set(self, name, value):
        self.value_map[name.lexeme] = value

        return value

    def __str__(self):
        return str(self.value_map)


#!/usr/bin/env python

from setuptools import setup, find_packages

version = '0.0.6'

setup(
    name='hyde-lang',
    author='Tyler Porter',
    author_email='tyler.b.porter@gmail.com',
    version=version,
    license='MIT',
    description='A Python-based interpretation of Lox',
    url='https://github.com/ty-porter/hyde',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
    keywords=[
        'hyde',
        'programming languages'
    ],
    platforms='ANY',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'hyde = hyde.__main__:run'
        ]
    }
)


from hyde.hyde_callable import HydeCallable
from hyde.hyde_instance import HydeInstance, HydeInstanceError


from http.server import BaseHTTPRequestHandler as PyBaseHttpRequestHandler
from http.server import HTTPServer
import threading
import os


class BasicHttpRequestHandler(HydeInstance):
    name = 'BasicHttpRequestHandler'

    class BasicHttpRequestHandlerFunction(HydeCallable):
        def __init__(self, handler, token):
            self.handler = handler
            self.token   = token

        
    class Serve(BasicHttpRequestHandlerFunction):
        arity = 0

        def call(self, _interpreter, _arguments):
            server = self._generate_singleton_http_server()
            thread = threading.Thread(
                target=server.serve_forever, kwargs={'poll_interval': 0.005}
            )
            thread.start()

        def _generate_singleton_http_server(self):
            request_attrs = self.handler.request_attrs

            class Handler(PyBaseHttpRequestHandler):
                def handle_single_request(self):
                    target_request_attr = None
                    content_type = self.guess_content_type()

                    if content_type == 'text/html':
                        for request_attr in request_attrs:
                            if request_attr['path'] == self.path:
                                target_request_attr = request_attr
                                break

                    if target_request_attr is None and content_type == 'text/html':
                        self.serve_404()
                    elif content_type != 'text/html':
                        self.serve_content(content_type)
                    else:
                        self.serve_html(request_attr['data'])

                def serve_404(self):
                    self.send_response(404)
                    self.send_header('content-type', 'text/html')
                    self.end_headers()
                    self.wfile.write('not found'.encode('utf-8'))

                def serve_content(self, content_type):
                    self.send_response(200)
                    self.send_header('content-type', content_type)
                    self.end_headers()

                    if os.path.exists(self.path[1:]):
                        with open(self.path[1:], 'rb') as content:
                            self.wfile.write(content.read())
                    else:
                        self.wfile.write(''.encode('utf-8'))

                def serve_html(self, html):
                    self.send_response(200)
                    self.send_header('content-type', 'text/html')
                    self.end_headers()
                    self.wfile.write(html.encode('utf-8'))

                def guess_content_type(self):
                    try:
                        if self.path[-4:] == '.css':
                            return 'text/css'
                        elif self.path[-4:] == '.png':
                            return 'image/png'
                        elif self.path[-5:] == '.woff':
                            return 'font/woff'
                        elif self.path[-4:] in ['.otf', '.ttf']:
                            return f'font/{self.path[-3:]}'

                        return 'text/html'
                    except Exception:
                        return 'text/html'

                def do_GET(self):
                    self.handle_single_request()

            port = int(os.environ.get('PORT', 5000))

            return HTTPServer(('', port), Handler)


    class ToString(BasicHttpRequestHandlerFunction):
        arity = 0

        def call(self, _interpreter, _arguments):
            return str(self.handler)


    def __init__(self, request_attrs):
        self.request_attrs = [element.value_map for element in request_attrs.elements]

    def get(self, name):
        if name.lexeme == 'serve':
            return BasicHttpRequestHandler.Serve(self, name)
        elif name.lexeme == 'toString':
            return BasicHttpRequestHandler.ToString(self, name)

        raise HydeInstanceError(name, f'Undefined property {name.lexeme} for {self.name} instance.')

    def set(self, name, _value):
        raise HydeInstanceError(name, "Can't add properties to BasicHttpRequestHandlers.")

    def __str__(self):
        return f'<BasicHttpRequestHandler route: {self.request_attrs}>'
